#!/usr/bin/env python
import json
import re
import pandas as pd
from Bio import SeqIO
from abromics_galaxy_json_extractor.interfaces import (
    AbromicsResultIterator,
)


required_metadata = {
    "hid": "Historic ID for plasmidfinder file from galaxy",
    "analysis_software_version": "plasmidfinder version",
    "reference_database_version": "plasmidfinder DB version",
    "plasmid_result_tabular_path": "plasmidfinder results in tabular format",
    "plasmid_result_tabular_hid": "plasmidfinder results hid in Galaxy",
    "genome_hit_path": (
        "fasta file with hits in genome, doesn't work for multiple input"
    ),
    "genome_hit_hid": "Historic ID for genome hit file from galaxy",
    "plasmid_hit_path": (
        "fasta file with plasmid sequences, doesn't work for multiple input"
    ),
    "plasmid_hit_hid": "Historic ID for plasmid sequence hit file from galaxy",
}


class PlasmidFinderIterator(AbromicsResultIterator):
    """Plasmidfinder tool parser
    Filter results from Plasmidfinder json result
    Could also parse file for:
    Plasmid sequence hit file
    Genome sequence hit
    Args:
        AbromicsResultIterator (abc.ABCMeta): abstract class to the parser
    """

    def __init__(self, source, metadata):
        metadata["analysis_software_name"] = "plasmidfinder"
        self.metadata = metadata
        super().__init__(source, self.metadata)
        self.abromics_results["tool_has_database"] = True
        self.parse()

    def parse(self):
        """Parse plasmidfinder file"""

        plasmid_infos = self._get_plasmid_results()

        if self.metadata["plasmid_result_tabular_path"]:
            self._check_input_files(
                self.metadata["plasmid_result_tabular_path"]
                )
            plasmid_df = self._get_mean_coverage(
                input_data=self.stream
                )
            plasmid_complete_df = self.get_plasmid_infos(
                plasmid_infos, plasmid_df
                )
            self.abromics_results["results"].append(
                {
                    "name": "plasmid_result_tabular",
                    "file_path": self.abromics_results.pop(
                        "plasmid_result_tabular_path"
                        ),
                    "hid": self.abromics_results.pop(
                        "plasmid_result_tabular_hid"
                        ),
                    "galaxy_id": "",
                    "content": "",
                }
            )
        else:
            plasmid_complete_df = plasmid_infos

        self.abromics_results["results"][0]["content"] = plasmid_complete_df

        if self.metadata["genome_hit_path"]:
            self._check_input_files(self.abromics_results["genome_hit_path"])
            extracted_infos = self.__get_fasta_infos__()
            self.abromics_results["results"].append(
                {
                    "name": "genome_hit",
                    "file_path": self.abromics_results.pop("genome_hit_path"),
                    "hid": self.abromics_results.pop("genome_hit_hid"),
                    "galaxy_id": "",
                    "content": [extracted_infos],
                }
            )

        if self.metadata["plasmid_hit_path"]:
            self._check_input_files(self.metadata["plasmid_hit_path"])
            extracted_infos = self.__get_fasta_infos__()
            self.abromics_results["results"].append(
                {
                    "name": "plasmid_hit",
                    "file_path": self.abromics_results.pop("plasmid_hit_path"),
                    "hid": self.abromics_results.pop("plasmid_hit_hid"),
                    "galaxy_id": "",
                    "content": [extracted_infos],
                }
            )

    def get_plasmid_infos(self, plasmid_dict, plasmid_df):
        plasmid_info_df = pd.DataFrame.from_dict(plasmid_dict)
        plasmid_info_df["group"] = (
                plasmid_info_df["gene_name"]
                + "_"
                + plasmid_info_df["positions_in_contig"]
            )
        plasmid_info_df = plasmid_info_df.merge(
                plasmid_df, how="inner", on=["group"]
            )
        del plasmid_info_df['group']
        plasmid_info_df = plasmid_info_df.to_dict(orient="records")
        return(plasmid_info_df)

    def _get_mean_coverage(self, input_data):
        plasmid_df = pd.read_table(self.stream, header=0, sep="\t")
        plasmid_df["group"] = (
            plasmid_df["Note"] + "_" + plasmid_df["Position in contig"]
        )
        # Extract column names of input dataframe
        filtered_df = self._filter_columns(
            plasmid_df,
            colnames="coverage|group")
        return filtered_df

    def _filter_columns(self, df, colnames="coverage|group"):
        # Extract column names of input dataframe
        df_colnames = list(df.columns)
        # Regex pattern for re.search: searching all names in cols
        # regex_query = r"|".join(cols)
        regex_query = colnames
        # Search for each column name in df_colnames and keep matched only
        filtered_columns = list(
            filter(lambda v: re.search(regex_query, v), df_colnames)
        )
        df = df[filtered_columns]
        return df

    def __get_fasta_infos__(self):
        """Extract fasta informations
        Extract name, sequences, positions
        and IS information
        Returns:
            dict: sequence informations
        """
        fasta_sequences = SeqIO.parse(self.stream, format="fasta")
        parsed_fasta_sequence = {}
        for seqname in fasta_sequences:
            sequence_infos = pd.DataFrame(seqname.description.split(" "))
            sequence_infos[[0, 1]] = sequence_infos[0].str.split(
                pat=":", n=1, expand=True
            )
            sequence_infos[[0, 1]] = sequence_infos[0].str.split(
                pat="=", n=1, expand=True
            )
            sequence_infos = sequence_infos.iloc[0:4, 0:2]
            sequence_infos = dict(
                zip(
                    sequence_infos[0],
                    sequence_infos[1],
                )
            )
            sequence_name = list(sequence_infos.keys())[0]
            sequence_infos.pop(sequence_name)
            parsed_fasta_sequence[sequence_name] = {
                "sequence_informations": sequence_infos,
                "nucleotide_sequence": seqname.seq.__str__(),
            }
        return parsed_fasta_sequence

    def _get_plasmid_results(self):
        json_object = json.loads(self.stream.read())
        json_object = json_object[list(json_object.keys())[0]].pop("results")
        plasmid_family_list = list(json_object.keys())
        results = [
            self._extract_plasmid_family_infos(
                json_input=json_object, plasmid_family=fam_name
            )
            for fam_name in plasmid_family_list
        ]
        clean_results = self._clean_plasmid_results(results)
        return clean_results

    def _clean_plasmid_results(self, raw_results):
        clean_results = []
        for i in range(len(raw_results)):
            for j in range(len(raw_results[i])):
                [
                    clean_results.append(k)
                    for k in raw_results[i][j]
                    if len(raw_results[i][j]) > 0
                ]
        return clean_results

    def _extract_plasmid_family_infos(self,
                                      json_input: dict,
                                      plasmid_family: str):
        plasmid_inc_list = list(json_input.get(plasmid_family).keys())
        return [
            self._extract_plasmid_incompatibility_infos(
                json_input=json_input,
                plasmid_group=plasmid_family,
                inc_name=inc
            )
            for inc in plasmid_inc_list
        ]

    def _extract_plasmid_incompatibility_infos(
        self, json_input: dict, plasmid_group: str, inc_name: str
    ):
        plasmid_hit = json_input.get(plasmid_group).get(inc_name)
        if isinstance(plasmid_hit, dict):
            return [
                self._extract_plasmid_infos(
                    contig_hit_name=key,
                    plasmid_hit_dict=plasmid_hit,
                    plasmid_group=plasmid_group,
                    inc_name=inc_name,
                )
                for key in plasmid_hit.keys()
            ]
        elif plasmid_hit == "No hit found":
            return []
        else:
            raise ValueError("Error when parsing plasmidfinder json results")

    def _extract_plasmid_infos(
        self,
        contig_hit_name,
        plasmid_hit_dict,
        plasmid_group: str,
        inc_name: str
    ):
        infos = plasmid_hit_dict.get(contig_hit_name)
        infos["inc_group"] = inc_name
        infos["ref_start"] = str.split(infos.get("position_in_ref"), "..")[0]
        infos["ref_end"] = str.split(infos.get("position_in_ref"), "..")[1]
        infos["contig_start"] = str.split(
            infos.get("positions_in_contig"), ".."
            )[0]
        infos["contig_end"] = str.split(
            infos.get("positions_in_contig"), ".."
            )[1]
        infos["contig_name"] = re.search(r"[^\s]+", contig_hit_name).group(0)
        infos["gene_name"] = infos.pop("note")
        infos["plasmid_group"] = plasmid_group
        return infos
